package org.xbib.jdbc.query;

import java.io.InputStream;
import java.nio.charset.CharsetDecoder;
import java.util.Properties;
import java.util.function.Function;
import java.util.function.Supplier;

/**
 * Pull configuration properties from various sources and filter/manipulate them.
 */
public interface ConfigSupplier extends Supplier<Config> {
    /**
     * Convenience method for fluent syntax.
     *
     * @return a builder for specifying from where configuration should be loaded
     */
    static ConfigSupplier of() {
        return new ConfigSupplierImpl();
    }

    static Config other(Function<String, String> other) {
        if (other instanceof Config) {
            return (Config) other;
        }
        return new ConfigSupplierImpl().custom(other::apply).get();
    }

    ConfigSupplier custom(Function<String, String> keyValueLookup);

    ConfigSupplier property(String key, String value);

    ConfigSupplier env();

    ConfigSupplier systemProperties();

    ConfigSupplier properties(Properties properties);

    ConfigSupplier config(Config config);

    ConfigSupplier config(Supplier<Config> config);

    /**
     * Adds a set of properties files to read from, which can be overridden by a system property "properties".
     * Equivalent to:
     * <pre>
     *   defaultPropertyFiles("properties", "conf/app.properties", "local.properties", "sample.properties")
     * </pre>
     */
    ConfigSupplier defaultProperties();

    /**
     * Adds a set of properties files to read from, which can be overridden by a specified system property.
     * Equivalent to:
     * <pre>
     *   defaultPropertyFiles(systemPropertyKey, Charset.defaultCharset().newDecoder(), filenames)
     * </pre>
     */
    ConfigSupplier defaultProperties(String systemPropertyKey, String... filenames);

    /**
     * Adds a set of properties files to read from, which can be overridden by a specified system property.
     * Equivalent to:
     * <pre>
     *   propertyFile(Charset.defaultCharset().newDecoder(),
     *       System.getProperty(systemPropertyKey, String.join(File.pathSeparator, filenames))
     *       .split(File.pathSeparator));
     * </pre>
     */
    ConfigSupplier defaultProperties(String systemPropertyKey, CharsetDecoder decoder, String... filenames);

    ConfigSupplier properties(String... filenames);

    ConfigSupplier properties(CharsetDecoder decoder, String... filenames);

    ConfigSupplier properties(InputStream... inputStreams);

    ConfigSupplier properties(CharsetDecoder decoder, InputStream... inputStreams);

    ConfigSupplier rename(String key, String newKey);

    ConfigSupplier includeKeys(String... keys);

    ConfigSupplier includePrefix(String... prefixes);

    ConfigSupplier includeRegex(String regex);

    ConfigSupplier excludeKeys(String... keys);

    ConfigSupplier excludePrefix(String... prefixes);

    ConfigSupplier excludeRegex(String regex);

    ConfigSupplier removePrefix(String... prefixes);

    ConfigSupplier addPrefix(String prefix);

    ConfigSupplier substitutions(Config config);

    Config get();
}
