package org.xbib.jdbc.query;

import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.nio.charset.Charset;
import java.nio.charset.CharsetDecoder;
import java.nio.file.FileSystems;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Properties;
import java.util.function.Function;
import java.util.function.Supplier;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * Access configuration properties from a variety of standard sources,
 * and provide some basic filtering and mapping of property keys.
 */
public class ConfigSupplierImpl implements ConfigSupplier {

    private static final String SEPARATOR = FileSystems.getDefault().getSeparator();

    private final List<Config> searchPath = new ArrayList<>();

    public ConfigSupplierImpl() {
        super();
    }

    public ConfigSupplierImpl(Config first) {
        searchPath.add(first);
    }

    @Override
    public ConfigSupplier custom(Function<String, String> keyValueLookup) {
        return custom(keyValueLookup, "custom()");
    }

    private ConfigSupplier custom(Function<String, String> keyValueLookup, String source) {
        searchPath.add(new ConfigImpl(keyValueLookup, source));
        return this;
    }

    @Override
    public ConfigSupplier property(String key, String value) {
        return custom(k -> k.equals(key) ? value : null, "value(" + key + ")");
    }

    @Override
    public ConfigSupplier config(Config config) {
        searchPath.add(config);
        return this;
    }

    @Override
    public ConfigSupplier config(Supplier<Config> config) {
        return config(config.get());
    }

    @Override
    public ConfigSupplier systemProperties() {
        return custom(System::getProperty, "systemProperties()");
    }

    @Override
    public ConfigSupplier env() {
        return custom(System::getenv, "env()");
    }

    @Override
    public ConfigSupplier properties(Properties properties) {
        return custom(properties::getProperty, "properties()");
    }

    @Override
    public ConfigSupplier defaultProperties() {
        return defaultProperties("properties", "conf/app.properties", "local.properties", "sample.properties");
    }

    @Override
    public ConfigSupplier defaultProperties(String systemPropertyKey, String... filenames) {
        return defaultProperties(systemPropertyKey, Charset.defaultCharset().newDecoder(), filenames);
    }

    @Override
    public ConfigSupplier defaultProperties(String systemPropertyKey, CharsetDecoder decoder, String... filenames) {
        String properties = System.getProperty(systemPropertyKey, String.join(SEPARATOR, filenames));
        return properties(Charset.defaultCharset().newDecoder(), properties.split(SEPARATOR));
    }

    @Override
    public ConfigSupplier properties(String... filenames) {
        return properties(Charset.defaultCharset().newDecoder(), filenames);
    }

    @Override
    public ConfigSupplier properties(CharsetDecoder decoder, String... filenames) {
        for (String filename : filenames) {
            if (filename != null) {
                Path path = Paths.get(filename);
                if (Files.exists(path)) {
                    try (InputStream inputStream = Files.newInputStream(path)) {
                        properties(decoder, inputStream);
                    } catch (IOException e) {
                        custom(k -> null, "Ignored: propertyFile(" + filename + ") " + e.getClass().getSimpleName());
                    }
                } else {
                    properties(decoder, getClass().getResourceAsStream(filename));
                }
            }
        }
        return this;
    }

    @Override
    public ConfigSupplier properties(InputStream... inputStreams) {
        return properties(Charset.defaultCharset().newDecoder(), inputStreams);
    }

    @Override
    public ConfigSupplier properties(CharsetDecoder decoder, InputStream... inputStreams) {
        for (InputStream inputStream : inputStreams) {
            Properties properties = new Properties();
            try (InputStreamReader reader = new InputStreamReader(inputStream, decoder)) {
                properties.load(reader);
                searchPath.add(new ConfigImpl(properties::getProperty, "propertyFile(" + inputStream + ")"));
            } catch (Exception e) {
                custom(k -> null, "Ignored: propertyFile(" + inputStream + ") " + e.getClass().getSimpleName());
            }
        }
        return this;
    }

    @Override
    public ConfigSupplier rename(String configKey, String newKey) {
        return new ConfigSupplierImpl(new ConfigImpl(key -> {
            if (key.equals(configKey)) {
                return null;
            }
            if (key.equals(newKey)) {
                return lookup(configKey);
            }
            return lookup(key);
        }, indentedSources("rename(" + configKey + " -> " + newKey + ")")));
    }

    @Override
    public ConfigSupplier includeKeys(String... keys) {
        return new ConfigSupplierImpl(new ConfigImpl(key -> {
            for (String k : keys) {
                if (key.equals(k)) {
                    return lookup(key);
                }
            }
            return null;
        }, indentedSources("includeKeys" + Arrays.asList(keys))));
    }

    @Override
    public ConfigSupplier includePrefix(String... prefixes) {
        return new ConfigSupplierImpl(new ConfigImpl(key -> {
            for (String prefix : prefixes) {
                if (key.startsWith(prefix)) {
                    return lookup(key);
                }
            }
            return null;
        }, indentedSources("includePrefix" + Arrays.asList(prefixes))));
    }

    @Override
    public ConfigSupplier includeRegex(String regex) {
        return new ConfigSupplierImpl(new ConfigImpl(key -> {
            if (key.matches(regex)) {
                return lookup(key);
            }
            return null;
        }, indentedSources("includeRegex(" + regex + ")")));
    }

    @Override
    public ConfigSupplier excludeKeys(String... keys) {
        return new ConfigSupplierImpl(new ConfigImpl(key -> {
            for (String k : keys) {
                if (key.equals(k)) {
                    return null;
                }
            }
            return lookup(key);
        }, indentedSources("excludeKeys" + Arrays.asList(keys))));
    }

    @Override
    public ConfigSupplier excludePrefix(String... prefixes) {
        return new ConfigSupplierImpl(new ConfigImpl(key -> {
            for (String prefix : prefixes) {
                if (key.startsWith(prefix)) {
                    return null;
                }
            }
            return lookup(key);
        }, indentedSources("excludePrefix" + Arrays.asList(prefixes))));
    }

    @Override
    public ConfigSupplier excludeRegex(String regex) {
        return new ConfigSupplierImpl(new ConfigImpl(key -> {
            if (key.matches(regex)) {
                return null;
            }
            return lookup(key);
        }, indentedSources("excludeRegex(" + regex + ")")));
    }

    @Override
    public ConfigSupplier removePrefix(String... prefixes) {
        return new ConfigSupplierImpl(new ConfigImpl(key -> {
            // Give precedence to ones that already lacked the prefix,
            // do an include*() first if you don't want that
            String value = lookup(key);
            if (value != null) {
                return value;
            }

            for (String prefix : prefixes) {
                value = lookup(prefix + key);
                if (value != null) {
                    return value;
                }
            }
            return null;
        }, indentedSources("removePrefix" + Arrays.asList(prefixes))));
    }

    @Override
    public ConfigSupplier addPrefix(String prefix) {
        return new ConfigSupplierImpl(new ConfigImpl(key -> {
            if (key.startsWith(prefix)) {
                return lookup(key.substring(prefix.length()));
            } else {
                return null;
            }
        }, indentedSources("addPrefix(" + prefix + ")")));
    }

    @Override
    public ConfigSupplier substitutions(Config config) {
        return new ConfigSupplierImpl(new ConfigImpl(key -> {
            String value = lookup(key);
            if (value != null) {
                // matches ${ENV_VAR_NAME} or $ENV_VAR_NAME
                Pattern p = Pattern.compile("(?<!\\$)\\$(?!\\$)\\{(\\w+)}|(?<!\\$)\\$(?!\\$)(\\w+)");
                Matcher m = p.matcher(value);
                StringBuffer sb = new StringBuffer();
                while (m.find()) {
                    String envVarName = null == m.group(1) ? m.group(2) : m.group(1);
                    String envVarValue = config.getString(envVarName);
                    m.appendReplacement(sb, null == envVarValue ? "" : envVarValue);
                }
                m.appendTail(sb);
                // Allow escaping literal $ with $$
                return sb.toString().replaceAll("(\\${2})", "\\$");
            } else {
                return null;
            }
        }, indentedSources("substitutions(" + config.sources() + ")")));
    }

    @Override
    public Config get() {
        return new ConfigImpl(this::lookup, indentedSources("Config"));
    }

    private String indentedSources(String label) {
        StringBuilder buf = new StringBuilder(label);
        for (Config config : searchPath) {
            buf.append(config.sources().replaceAll("(?s)^|\\n", "\n  "));
        }
        return buf.toString();
    }

    private String lookup(String key) {
        for (Config config : searchPath) {
            String value = config.getString(key);
            if (value != null) {
                return value;
            }
        }
        return null;
    }
}
