package org.xbib.jdbc.query;

import java.sql.Connection;
import java.sql.SQLException;
import java.util.List;
import java.util.Map;
import java.util.function.Consumer;
import java.util.function.Supplier;

/**
 * Primary class for accessing a relational (SQL) database.
 */
public interface Database extends Supplier<Database> {

    int jdbcMajorVersion();

    int jdbcMinorVersion();

    /**
     * Create a SQL "insert" statement for further manipulation and execution.
     * Note this call does not actually execute the SQL.
     *
     * @param sql the SQL to execute, optionally containing indexed ("?") or
     *            named (":foo") parameters. To include the characters '?' or ':'
     *            in the SQL you must escape them with two ("??" or "::"). You
     *            MUST be careful not to pass untrusted strings in as SQL, since
     *            this will be executed in the database.
     * @return an interface for further manipulating the statement; never null
     */
    SqlInsert toInsert(String sql);

    SqlInsert toInsert(Sql sql);

    /**
     * Create a SQL "select" statement for further manipulation and execution.
     * Note this call does not actually execute the SQL.
     *
     * @param sql the SQL to execute, optionally containing indexed ("?") or
     *            named (":foo") parameters. To include the characters '?' or ':'
     *            in the SQL you must escape them with two ("??" or "::"). You
     *            MUST be careful not to pass untrusted strings in as SQL, since
     *            this will be executed in the database.
     * @return an interface for further manipulating the statement; never null
     */
    SqlSelect toSelect(String sql);

    SqlSelect toSelect(Sql sql);

    /**
     * Create a SQL "update" statement for further manipulation and execution.
     * Note this call does not actually execute the SQL.
     *
     * @param sql the SQL to execute, optionally containing indexed ("?") or
     *            named (":foo") parameters. To include the characters '?' or ':'
     *            in the SQL you must escape them with two ("??" or "::"). You
     *            MUST be careful not to pass untrusted strings in as SQL, since
     *            this will be executed in the database.
     * @return an interface for further manipulating the statement; never null
     */
    SqlUpdate toUpdate(String sql);

    SqlUpdate toUpdate(Sql sql);

    /**
     * Create a SQL "delete" statement for further manipulation and execution.
     * Note this call does not actually execute the SQL.
     *
     * @param sql the SQL to execute, optionally containing indexed ("?") or
     *            named (":foo") parameters. To include the characters '?' or ':'
     *            in the SQL you must escape them with two ("??" or "::"). You
     *            MUST be careful not to pass untrusted strings in as SQL, since
     *            this will be executed in the database.
     * @return an interface for further manipulating the statement; never null
     */
    SqlUpdate toDelete(String sql);

    SqlUpdate toDelete(Sql sql);

    /**
     * Create a DDL (schema modifying) statement for further manipulation and execution.
     * Note this call does not actually execute the SQL.
     *
     * @param sql the SQL to execute, optionally containing indexed ("?") or
     *            named (":foo") parameters. To include the characters '?' or ':'
     *            in the SQL you must escape them with two ("??" or "::"). You
     *            MUST be careful not to pass untrusted strings in as SQL, since
     *            this will be executed in the database.
     * @return an interface for further manipulating the statement; never null
     */
    Ddl ddl(String sql);

    /**
     * Read the next value from a sequence. This method helps smooth over the
     * syntax differences across databases.
     */
    Long nextSequenceValue(String sequenceName);

    /**
     * Cause the underlying connection to commit its transaction immediately. This
     * must be explicitly enabled (see {@link Options},
     * or it will throw a {@link DatabaseException}.
     */
    void commitNow();

    /**
     * Cause the underlying connection to roll back its transaction immediately. This
     * must be explicitly enabled (see {@link Options},
     * or it will throw a {@link DatabaseException}.
     */
    void rollbackNow();

    /**
     * <p>Obtain direct access to the connection being used by this instance. Be very
     * careful as this is highly likely to be unsafe and cause you great pain and
     * suffering. This method is included to help ease into the library in large
     * codebases where some parts still rely on direct JDBC access.</p>
     *
     * <p>By default this method will throw a {@link DatabaseException}. If you want
     * to use this method you must explicitly enable it via
     * {@link Options#allowConnectionAccess()}</p>
     */
    Connection underlyingConnection();

    /**
     * Probe the underlying connection.
     * @return true if database has executed a probe query successfully, false if not
     */
    boolean probe(int seconds) throws SQLException;

    Options options();

    /**
     * Access information about what kind of database we are dealing with.
     */
    Flavor flavor();

    /**
     * <p>A little syntax sugar to make it easier to customize your SQL based on the
     * specific database. For example:</p>
     *
     * <pre>"select 1" + db.when().oracle(" from dual")</pre>
     * <pre>"select " + db.when().postgres("date_trunc('day',").other("trunc(") + ") ..."</pre>
     *
     * @return an interface for chaining or terminating the conditionals
     */
    When when();

    /**
     * Convenience method to deal with mutually incompatible syntax for this. For example:
     *
     * <p>Oracle: 'drop sequence x'</p>
     * <p>Derby: 'drop sequence x restrict'</p>"
     */
    void dropSequenceQuietly(String sequenceName);

    /**
     * Convenience method to deal with dropping tables that may or may not exist. Some
     * databases make it hard to check and conditionally drop things, so we will just
     * try to drop it and ignore the errors.
     *
     * @param tableName the table to be dropped
     */
    void dropTableQuietly(String tableName);

    /**
     * Convenience method to check if a table or view exists so that caller can decide
     * whether to create or update a table. The table name's case is normalized using
     * the database's convention unless tableName is enclosed in double quotes.
     * The default catalog and schema from the DB connection will be used.
     *
     * @param tableName the table to be checked
     * @return true if the table or view exists
     */
    boolean tableExists(String tableName);

    /**
     * Convenience method to check whether a table or view exists or not.
     * The table name's case is normalized using the database's convention
     * unless tableName is enclosed in double quotes.  The default catalog
     * from the DB connection will be used.
     *
     * @param tableName  the table to be checked
     * @param schemaName the schema expected to contain the table
     * @return true if the table or view exists
     */
    boolean tableExists(String tableName, String schemaName);

    /**
     * Convenience method to get all column sizes from a table.
     * @param tableName the table to be checked
     * @return a map of columns names with size
     */
    Map<String, Integer> getColumnSizesOfTable(String tableName);

    /**
     * Check the JVM time (and timezone) against the database and log a warning
     * or throw an error if they are too far apart. It is a good idea to do this
     * before you store and dates, and maybe make it part of your health checks.
     * If the clocks differ by more than an hour, a DatabaseException is thrown
     * suggesting you check the timezones (under the assumptions the JVM and
     * database are running in different timezones).
     *
     * @param millisToWarn  if the clocks disagree by more than this and less than
     *                      millisToError, a warning will be dropped in the log
     * @param millisToError if the clocks disagree by more than this a
     *                      DatabaseEception will be thrown
     */
    void assertTimeSynchronized(long millisToWarn, long millisToError);

    /**
     * Convenience method, same as {@link #assertTimeSynchronized(long, long)}
     * with millisToWarn=10000 and millisToError=30000.
     */
    void assertTimeSynchronized();

    Table getPagedRows(Table table);

    long getRowsCount(Table table);

    long countRows(String statement, Map<String, Object> params);

    Table getTable(String statement, Map<String, Object> params);

    Table getSingleRow(String statement, Map<String, Object> params);

    Table getTable(String statement,
                   Map<String, Object> params,
                   int limit,
                   int fetchSize,
                   int timeoutSeconds);

    void consumeRows(String statement,
                     Map<String, Object> params,
                     int limit,
                     int fetchSize,
                     int timeoutSeconds,
                     Consumer<List<Object>> consumer);

    void insert(String statement, Map<String, Object> params);

    void insert(String statement, List<Map<String, Object>> params);

    void update(String statement, Map<String, Object> params);

    void update(String statement, List<Map<String, Object>> params);

    void upsert(String insertStatement, String updateStatement, Map<String, Object> params);

    void delete(String statement, Map<String, Object> params);

    void delete(String statement, List<Map<String, Object>> params);

    /**
     * Write data into a queue table, select a channel, and consume the returned primary keys.
     *
     * @param table the queue table
     * @param channel the queue table channel
     * @param data the data for the queue table
     * @param consumer the consumer for the primary keys
     * @throws SQLException if writing fails
     */
    void writeQueue(String table, String channel, String data, Consumer<Long> consumer) throws SQLException;

    /**
     * Proble queue for length.
     * @param table the queue table
     * @param channel the queue table channel. A queue table can have many channels
     * @return the queue length
     * @throws SQLException if statement fails
     */
    long probeQueue(String table, String channel) throws SQLException;

    /**
     * Read from a queue table until limit is reached and lock rows in a transaction until processing by a consumer
     * has succeeded or failed.
     * Commit only if processing was successful, otherwise roll back the transaction.
     * The connection is switched into manual commit and afterwards, auto-commit is enabled again.
     * @param table the queue table
     * @param channel the queue table channel. A queue table can have many channels
     * @param consumer a consumer for the queue table data column or null
     * @throws SQLException if rollback fails
     */
    void consumeQueue(String table, String channel, int limit, Consumer<String> consumer) throws SQLException;
}
