package org.xbib.jdbc.query;

import java.util.function.Supplier;

/**
 * This builder is immutable, so setting various options does not affect
 * the previous instance. This is intended to make it safe to pass builders
 * around without risk someone will reconfigure it.
 */
public interface DatabaseProviderBuilder {

    DatabaseProviderBuilder withOptions(OptionsOverride options);

    /**
     * Enable logging of parameter values along with the SQL.
     */

    DatabaseProviderBuilder withSqlParameterLogging();

    /**
     * Include SQL in exception messages. This will also include parameters in the
     * exception messages if SQL parameter logging is enabled. This is handy for
     * development, but be careful as this is an information disclosure risk,
     * dependent on how the exception are caught and handled.
     */

    DatabaseProviderBuilder withSqlInExceptionMessages();

    /**
     * Wherever argDateNowPerDb() is specified, use argDateNowPerApp() instead. This is
     * useful for testing purposes as you can use OptionsOverride to provide your
     * own system clock that will be used for time travel.
     */

    DatabaseProviderBuilder withDatePerAppOnly();

    /**
     * Allow provided Database instances to explicitly control transactions using the
     * commitNow() and rollbackNow() methods. Otherwise calling those methods would
     * throw an exception.
     */

    DatabaseProviderBuilder withTransactionControl();

    /**
     * This can be useful when testing code, as it can pretend to use transactions,
     * while giving you control over whether it actually commits or rolls back.
     */

    DatabaseProviderBuilder withTransactionControlSilentlyIgnored();

    /**
     * Allow direct access to the underlying database connection. Normally this is
     * not allowed, and is a bad idea, but it can be helpful when migrating from
     * legacy code that works with raw JDBC.
     */

    DatabaseProviderBuilder withConnectionAccess();

    /**
     * WARNING: You should try to avoid using this method. If you use it more
     * that once or twice in your entire codebase you are probably doing
     * something wrong.
     *
     * <p>If you use this method you are responsible for managing
     * the transaction and commit/rollback/close.</p>
     */

    DatabaseProvider build();

    /**
     * This is a convenience method to eliminate the need for explicitly
     * managing the resources (and error handling) for this class. After
     * the run block is complete the transaction will commit unless the
     * {@link DbCode#run(Supplier) run(Supplier)} method threw a {@link Throwable}.
     *
     * <p>Here is a typical usage:
     * <pre>
     *   dbp.transact(dbs -&gt; {
     *     List<String> r = dbs.get().toSelect("select a from b where c=?").argInteger(1).queryStrings();
     *     ... do something with the results ...
     *   });
     * </pre>
     * </p>
     *
     * @param code the code you want to run as a transaction with a Database
     * @see #transact(DbCodeTx)
     */
    void transact(DbCode code);

    /**
     * This method is the same as {@link #transact(DbCode)} but allows a return value.
     *
     * <p>Here is a typical usage:
     * <pre>
     *   List<String> r = dbp.transact(dbs -&gt; {
     *     return dbs.get().toSelect("select a from b where c=?").argInteger(1).queryStrings();
     *   });
     * </pre>
     * </p>
     */
    <T> T transactReturning(DbCodeTyped<T> code);

    /**
     * This is a convenience method to eliminate the need for explicitly
     * managing the resources (and error handling) for this class. After
     * the run block is complete commit() will be called unless either the
     * {@link DbCodeTx#run(Supplier, Transaction)} method threw a {@link Throwable}
     * while {@link Transaction#isRollbackOnError()} returns true, or
     * {@link Transaction#isRollbackOnly()} returns a true value.
     *
     * <p>Here is a typical usage:
     * <pre>
     *   dbp.transact((dbs, tx) -&gt; {
     *     tx.setRollbackOnError(false);
     *     dbs.get().toInsert("...").argInteger(1).insert(1);
     *     ...some stuff that might fail...
     *   });
     * </pre>
     * </p>
     *
     * @param code the code you want to run as a transaction with a Database
     */
    void transact(DbCodeTx code);

}
