package org.xbib.jdbc.query;

import java.sql.Statement;
import java.sql.Types;
import org.xbib.jdbc.query.util.DebugSql;
import org.xbib.jdbc.query.util.InternalStringReader;
import org.xbib.jdbc.query.util.Metric;
import org.xbib.jdbc.query.util.RewriteArg;

import java.io.InputStream;
import java.io.Reader;
import java.math.BigDecimal;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.time.LocalDate;
import java.time.LocalDateTime;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * This is the key class for configuring (query parameters) and executing a database query.
 */
public class SqlUpdateImpl implements SqlUpdate {
    
    private static final Logger log = Logger.getLogger(Database.class.getName());
    
    private final Connection connection;
    
    private final StatementAdapter adaptor;
    
    private final String sql;
    
    private final Options options;

    private List<Batch> batched;

    private List<Object> parameterList;
    
    private Map<String, Object> parameterMap;
    
    public SqlUpdateImpl(Connection connection, String sql, Options options) {
        this.connection = connection;
        this.sql = sql;
        this.options = options;
        adaptor = new StatementAdapter(options);
    }
    
    @Override
    public SqlUpdate argBoolean(Boolean arg) {
        return positionalArg(adaptor.nullString(booleanToString(arg)));
    }

    @Override
    public SqlUpdate argBoolean(String argName, Boolean arg) {
        return namedArg(argName, adaptor.nullString(booleanToString(arg)));
    }

    @Override
    public SqlUpdate argInteger(Integer arg) {
        return positionalArg(adaptor.nullNumeric(arg));
    }

    @Override
    public SqlUpdate argInteger(String argName, Integer arg) {
        return namedArg(argName, adaptor.nullNumeric(arg));
    }

    @Override
    public SqlUpdate argLong(Long arg) {
        return positionalArg(adaptor.nullNumeric(arg));
    }

    @Override
    public SqlUpdate argLong(String argName, Long arg) {
        return namedArg(argName, adaptor.nullNumeric(arg));
    }

    @Override
    public SqlUpdate argFloat(Float arg) {
        return positionalArg(adaptor.nullNumeric(arg));
    }

    @Override
    public SqlUpdate argFloat(String argName, Float arg) {
        return namedArg(argName, adaptor.nullNumeric(arg));
    }

    @Override
    public SqlUpdate argDouble(Double arg) {
        return positionalArg(adaptor.nullNumeric(arg));
    }

    @Override
    public SqlUpdate argDouble(String argName, Double arg) {
        return namedArg(argName, adaptor.nullNumeric(arg));
    }

    @Override
    public SqlUpdate argBigDecimal(BigDecimal arg) {
        return positionalArg(adaptor.nullNumeric(arg));
    }

    @Override
    public SqlUpdate argBigDecimal(String argName, BigDecimal arg) {
        return namedArg(argName, adaptor.nullNumeric(arg));
    }

    @Override
    public SqlUpdate argString(String arg) {
        return positionalArg(adaptor.nullString(arg));
    }

    @Override
    public SqlUpdate argString(String argName, String arg) {
        return namedArg(argName, adaptor.nullString(arg));
    }

    @Override
    public SqlUpdate argLocalDateTime(LocalDateTime arg) {
        return positionalArg(adaptor.nullLocalDateTime(arg));
    }

    @Override
    public SqlUpdate argLocalDateTime(String argName, LocalDateTime arg) {
        return namedArg(argName, adaptor.nullLocalDateTime(arg));
    }

    @Override
    public SqlUpdate argLocalDate(LocalDate arg) {
        return positionalArg(adaptor.nullLocalDate(arg));
    }

    @Override
    public SqlUpdate argLocalDate(String argName, LocalDate arg) {
        return namedArg(argName, adaptor.nullLocalDate(arg));
    }

    @Override
    public SqlUpdate argLocalDateTimeNowPerDb() {
        if (options.useLocalDateTimeOnly()) {
            return positionalArg(adaptor.nullLocalDateTime(LocalDateTime.now()));
        }
        return positionalArg(new RewriteArg(options.flavor().dbTimeMillis()));
    }

    @Override
    public SqlUpdate argLocalDateTimeNowPerDb(String argName) {
        if (options.useLocalDateTimeOnly()) {
            return namedArg(argName, adaptor.nullLocalDateTime(LocalDateTime.now()));
        }
        return namedArg(argName, new RewriteArg(options.flavor().dbTimeMillis()));
    }

    @Override
    public SqlUpdate argBlobBytes(byte[] arg) {
        return positionalArg(adaptor.nullBytes(arg));
    }

    @Override
    public SqlUpdate argBlobBytes(String argName, byte[] arg) {
        return namedArg(argName, adaptor.nullBytes(arg));
    }

    @Override
    public SqlUpdate argBlobStream(InputStream arg) {
        return positionalArg(adaptor.nullInputStream(arg));
    }

    @Override
    public SqlUpdate argBlobStream(String argName, InputStream arg) {
        return namedArg(argName, adaptor.nullInputStream(arg));
    }

    @Override
    public SqlUpdate argClobString(String arg) {
        return positionalArg(adaptor.nullClobReader(arg == null ? null : new InternalStringReader(arg)));
    }

    @Override
    public SqlUpdate argClobString(String argName, String arg) {
        return namedArg(argName, adaptor.nullClobReader(arg == null ? null : new InternalStringReader(arg)));
    }

    @Override
    public SqlUpdate argClobReader(Reader arg) {
        return positionalArg(adaptor.nullClobReader(arg));
    }

    @Override
    public SqlUpdate argClobReader(String argName, Reader arg) {
        return namedArg(argName, adaptor.nullClobReader(arg));
    }

    @Override
    public SqlUpdate argNull(String argName) {
        return namedArg(argName, new SqlNull(Types.VARCHAR));
    }

    @Override
    public SqlUpdate withArgs(SqlArgs args) {
        return apply(args);
    }

    @Override
    public SqlUpdate apply(Apply apply) {
        apply.apply(this);
        return this;
    }

    @Override
    public SqlUpdate batch() {
        if ((parameterList != null && !parameterList.isEmpty())
                || (parameterMap != null && !parameterMap.isEmpty())) {
            if (batched == null) {
                batched = new ArrayList<>();
            }
            batched.add(new Batch(parameterList, parameterMap));
            parameterList = new ArrayList<>();
            parameterMap = new HashMap<>();
        }
        return this;
    }

    @Override
    public int update() {
        return updateInternal(0);
    }

    @Override
    public void update(int expectedNumAffectedRows) {
        updateInternal(expectedNumAffectedRows);
    }

    @Override
    public void updateBatch() {
        int[] result = updateBatchInternal();
        for (int r : result) {
            // Tolerate SUCCESS_NO_INFO for older versions of Oracle
            if (r != 1 && r != Statement.SUCCESS_NO_INFO) {
                throw new DatabaseException("Batch did not return the expected result: " + Arrays.toString(result));
            }
        }
    }

    @Override
    public int[] updateBatchUnchecked() {
        return updateBatchInternal();
    }

    private int[] updateBatchInternal() {
        batch();
        if (batched == null || batched.size() == 0) {
            throw new DatabaseException("Batch insert requires parameters");
        }
        PreparedStatement ps = null;
        Metric metric = new Metric(log.isLoggable(Level.FINE));
        String executeSql = sql;
        Object[] firstRowParameters = null;
        List<Object[]> parameters = new ArrayList<>();
        boolean isSuccess = false;
        String errorCode = null;
        Exception logEx = null;
        try {
            for (Batch batch : batched) {
                MixedParameterSql mpSql = new MixedParameterSql(sql, batch.parameterList, batch.parameterMap);
                if (firstRowParameters == null) {
                    executeSql = mpSql.getSqlToExecute();
                    firstRowParameters = mpSql.getArgs();
                } else {
                    if (!executeSql.equals(mpSql.getSqlToExecute())) {
                        throw new DatabaseException("All rows in a batch must use parameters in the same way. \nSQL1: "
                                + executeSql + "\nSQL2: " + mpSql.getSqlToExecute());
                    }
                }
                parameters.add(mpSql.getArgs());
            }
            if (connection != null) {
                ps = connection.prepareStatement(executeSql);
                for (Object[] params : parameters) {
                    adaptor.addParameters(ps, params);
                    ps.addBatch();
                }
                metric.checkpoint("prep");
                int[] numAffectedRows = ps.executeBatch();
                metric.checkpoint("execBatch", parameters.size());
                isSuccess = true;
                return numAffectedRows;
            } else {
                return new int[]{-1};
            }
        } catch (WrongNumberOfRowsException e) {
            throw e;
        } catch (Exception e) {
            errorCode = options.generateErrorCode();
            logEx = e;
            throw DatabaseException.wrap(DebugSql.exceptionMessage(executeSql, firstRowParameters, errorCode, options), e);
        } finally {
            adaptor.closeQuietly(ps, log);
            metric.done("close");
            if (isSuccess) {
                DebugSql.logSuccess("Insert", log, metric, executeSql, firstRowParameters, options);
            } else {
                DebugSql.logError("Insert", log, metric, errorCode, executeSql, firstRowParameters, options, logEx);
            }
        }
    }

    private int updateInternal(int expectedNumAffectedRows) {
        PreparedStatement ps = null;
        Metric metric = new Metric(log.isLoggable(Level.FINE));
        String executeSql = sql;
        Object[] parameters = null;
        boolean isSuccess = false;
        String errorCode = null;
        Exception logEx = null;
        try {
            MixedParameterSql mpSql = new MixedParameterSql(sql, parameterList, parameterMap);
            executeSql = mpSql.getSqlToExecute();
            parameters = mpSql.getArgs();
            if (connection != null) {
                ps = connection.prepareStatement(executeSql);
                adaptor.addParameters(ps, parameters);
                metric.checkpoint("prep");
                int numAffectedRows = ps.executeUpdate();
                metric.checkpoint("exec", numAffectedRows);
                if (expectedNumAffectedRows > 0 && numAffectedRows != expectedNumAffectedRows) {
                    errorCode = options.generateErrorCode();
                    throw new WrongNumberOfRowsException("The number of affected rows was " + numAffectedRows + ", but "
                            + expectedNumAffectedRows + " were expected." + "\n"
                            + DebugSql.exceptionMessage(executeSql, parameters, errorCode, options));
                }
                isSuccess = true;
                return numAffectedRows;
            } else {
                return -1;
            }
        } catch (WrongNumberOfRowsException e) {
            throw e;
        } catch (Exception e) {
            errorCode = options.generateErrorCode();
            logEx = e;
            throw DatabaseException.wrap(DebugSql.exceptionMessage(executeSql, parameters, errorCode, options), e);
        } finally {
            adaptor.closeQuietly(ps, log);
            metric.done("close");
            if (isSuccess) {
                DebugSql.logSuccess("Update", log, metric, executeSql, parameters, options);
            } else {
                DebugSql.logError("Update", log, metric, errorCode, executeSql, parameters, options, logEx);
            }
        }
    }
    
    private SqlUpdate positionalArg(Object arg) {
        if (parameterList == null) {
            parameterList = new ArrayList<>();
        }
        parameterList.add(arg);
        return this;
    }
    
    private SqlUpdate namedArg(String argName, Object arg) {
        if (parameterMap == null) {
            parameterMap = new HashMap<>();
        }
        if (argName.startsWith(":")) {
            argName = argName.substring(1);
        }
        parameterMap.put(argName, arg);
        return this;
    }

    private String booleanToString(Boolean b) {
        return b == null ? null : b ? "Y" : "N";
    }

    private static class Batch {
        private final List<Object> parameterList;
        private final Map<String, Object> parameterMap;

        public Batch(List<Object> parameterList, Map<String, Object> parameterMap) {
            this.parameterList = parameterList;
            this.parameterMap = parameterMap;
        }
    }
}
